@@ -1,8 +1,10 @@
+#![cfg_attr(feature = "auto-sequence", feature(proc_macro_span))]
+
 use nom::{
     branch::alt,
-    combinator::{map, opt},
+    combinator::{map, opt, verify},
     error::{make_error, ErrorKind},
-    multi::{many0, many1},
+    multi::many0,
     sequence::tuple,
     IResult,
 };
@@ -19,11 +21,11 @@ pub fn rule(tokens: proc_macro::TokenStream) -> proc_macro::TokenStream {
     let i: Vec<TokenTree> = tokens.into_iter().collect();
 
     let (i, (match_text, _, match_token, _)) =
-        tuple((group, match_punct(','), group, match_punct(',')))(&i).unwrap();
+        tuple((path, match_punct(','), path, match_punct(',')))(&i).unwrap();
 
     let terminal = CustomTerminal {
-        match_text,
-        match_token,
+        match_text: match_text.1,
+        match_token: match_token.1,
     };
 
     let rule = parse_rule(i.iter().cloned().collect());
@@ -82,8 +84,8 @@ enum ReturnType {
 }
 
 struct CustomTerminal {
-    match_text: Group,
-    match_token: Group,
+    match_text: Path,
+    match_token: Path,
 }
 
 type Input<'a> = &'a [TokenTree];
@@ -148,16 +150,17 @@ fn path<'a>(i: Input<'a>) -> IResult<Input<'a>, (Span, Path)> {
 fn parse_rule(tokens: TokenStream) -> Rule {
     let i: Vec<TokenTree> = tokens.into_iter().collect();
 
-    let (i, elems) = many1(parse_rule_element)(&i).unwrap();
-
+    let (i, elems) = many0(parse_rule_element)(&i).unwrap();
     if !i.is_empty() {
         let rest: TokenStream = i.iter().cloned().collect();
         abort!(rest, "unable to parse the following rules: {}", rest);
     }
 
+    #[cfg(feature = "auto-sequence")]
+    let elems = auto_sequence::amend_sequence(elems);
+
     let mut iter = elems.into_iter().peekable();
     let rule = unwrap_pratt(RuleParser.parse(&mut iter));
-
     if iter.peek().is_some() {
         let rest: Vec<_> = iter.collect();
         abort!(
@@ -171,20 +174,30 @@ fn parse_rule(tokens: TokenStream) -> Rule {
 }
 
 fn parse_rule_element<'a>(i: Input<'a>) -> IResult<Input<'a>, WithSpan> {
-    let function_call = map(
-        tuple((match_punct('#'), path, opt(group))),
-        |(hashtag, (path_span, fn_path), args)| {
-            let span = hashtag.span().join(path_span).unwrap();
-            let span = args
-                .as_ref()
-                .map(|args| args.span().join(span).unwrap())
-                .unwrap_or(span);
+    let function_call = |i| {
+        let (i, hashtag) = match_punct('#')(i)?;
+        let (i, (path_span, fn_path)) = path(i)?;
+        let (i, args) = opt(verify(group, |g| {
+            if cfg!(feature = "auto-sequence") {
+                path_span.end() == g.span().start()
+            } else {
+                true
+            }
+        }))(i)?;
+        let span = hashtag.span().join(path_span).unwrap();
+        let span = args
+            .as_ref()
+            .map(|args| args.span().join(span).unwrap())
+            .unwrap_or(span);
+
+        Ok((
+            i,
             WithSpan {
                 elem: RuleElement::ExternalFunction(fn_path, args),
                 span,
-            }
-        },
-    );
+            },
+        ))
+    };
     let context = map(tuple((match_punct(':'), literal)), |(colon, msg)| {
         let span = colon.span().join(msg.span()).unwrap();
         WithSpan {
@@ -238,6 +251,48 @@ fn parse_rule_element<'a>(i: Input<'a>) -> IResult<Input<'a>, WithSpan> {
     ))(i)
 }
 
+#[cfg(feature = "auto-sequence")]
+mod auto_sequence {
+    use super::*;
+
+    // Automatically insert `RuleElement::Sequence` between primaries.
+    pub(crate) fn amend_sequence(elems: impl IntoIterator<Item = WithSpan>) -> Vec<WithSpan> {
+        let mut output = Vec::new();
+        let mut iter = elems.into_iter().peekable();
+        loop {
+            match (iter.next(), iter.peek()) {
+                (Some(lhs), Some(rhs)) if should_amend(&lhs, rhs) => {
+                    let span = lhs.span.join(rhs.span).unwrap();
+                    output.push(lhs);
+                    output.push(WithSpan {
+                        elem: RuleElement::Sequence,
+                        span,
+                    });
+                }
+                (Some(elem), _) => {
+                    output.push(elem);
+                }
+                (None, _) => break,
+            }
+        }
+        output
+    }
+
+    fn should_amend(lhs: &WithSpan, rhs: &WithSpan) -> bool {
+        match (query_affix(lhs), query_affix(rhs)) {
+            (Affix::Nilfix, Affix::Nilfix)
+            | (Affix::Nilfix, Affix::Prefix(_))
+            | (Affix::Postfix(_), Affix::Nilfix)
+            | (Affix::Postfix(_), Affix::Prefix(_)) => true,
+            _ => false,
+        }
+    }
+
+    fn query_affix(elem: &WithSpan) -> Affix {
+        PrattParser::<std::iter::Once<_>>::query(&mut RuleParser, elem).unwrap()
+    }
+}
+
 fn unwrap_pratt(res: Result<Rule, PrattError<WithSpan, pratt::NoError>>) -> Rule {
     match res {
         Ok(res) => res,
