@@ -8,7 +8,7 @@ use nom::{
 use TokenKind::*;
 
 macro_rules! rule {
-    ($($tt:tt)*) => { nom_rule::rule!((match_text), (match_token), $($tt)*) }
+    ($($tt:tt)*) => { nom_rule::rule!(match_text, match_token, $($tt)*) }
 }
 
 #[test]
@@ -65,6 +65,32 @@ fn sql_create_table() {
     );
 }
 
+#[cfg(feature = "auto-sequence")]
+#[test]
+fn sql_column_ref() {
+    let tokens = tokenise("a.b");
+
+    let mut rule = rule!(
+         #ident ("." #ident)?
+    );
+
+    let res: IResult<_, _> = rule(&tokens);
+    assert_eq!(
+        dbg!(res.unwrap().1),
+        (
+            "a",
+            Some((
+                &Token {
+                    kind: Whitespace,
+                    text: ".",
+                    span: 1..2,
+                },
+                "b",
+            ),),
+        ),
+    );
+}
+
 #[derive(Logos, Clone, Copy, Debug, PartialEq)]
 enum TokenKind {
     #[error]
